# coding: utf-8
"""
pronto.parser.utils
=================

This module defines mapping to convert metadata from obo to owl and owl to
obo, as well as enums to state the section of the ontology the parser is
currently looking at.
"""
from __future__ import unicode_literals

import six

try:                        # Use enums if possible to improve
    from enum import Enum   # output but don't have to dl enum32
except ImportError:         # backport as it's not that important
    Enum = object           # enough to truly depend on it


class OboSection(Enum):
    meta    = 1
    typedef = 2
    term    = 3

class OwlSection(Enum):
    ontology = 1
    classes  = 2


owl_ns = {
    'dc':       "http://purl.org/dc/elements/1.1/",
    'doap':     "http://usefulinc.com/ns/doap#",
    'foaf':     "http://xmlns.com/foaf/0.1/",
    'meta':     "http://www.co-ode.org/ontologies/meta.owl#",
    'obo':      "http://purl.obolibrary.org/obo/",
    'oboInOwl': "http://www.geneontology.org/formats/oboInOwl#",
    'owl':      "http://www.w3.org/2002/07/owl#",
    'protege':  "http://protege.stanford.edu/plugins/owl/protege#",
    'rdf':      "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    'rdfs':     "http://www.w3.org/2000/01/rdf-schema#",
    'skos':     "http://www.w3.org/2004/02/skos/core#",
    'ubprop':   "http://purl.obolibrary.org/obo/ubprop#",
    'uberon':   "http://purl.obolibrary.org/obo/uberon#",
    'xsd':      "http://www.w3.org/2001/XMLSchema#",
}

owl_to_obo = {
    'hasDbXref': 'xref',
    'equivalentClass': 'equivalent_to',
    'inSubset': 'subset',
    'hasOBONamespace': 'namespace',
    'hasOBOFormatVersion': 'format-version',
    'hasExactSynonym': 'exact_synonym',
    'hasBroadSynonym': 'broad_synonym',
    'hasNarrowSynonym': 'narrow_synonym',

    #FEAT# Translate Information Ontology
    #FEAT# 'IAO_0000115': 'definition',

    #FEAT# Extract Owl defined Relationship
    #FEAT# 'is_metadata_tag': 'is_metadata_tag',
}

obo_to_owl = {v:k for k,v in six.iteritems(owl_to_obo)}



# coding: utf-8
"""
**pronto**: a Python frontend to ontologies
===========================================

"""

from __future__ import absolute_import

__all__ = ["Ontology", "Term", "TermList", "Relationship"]
__version__='0.6.0'
__author__='Martin Larralde'
__author_email__ = 'martin.larralde@ens-cachan.fr'


try:
    from .ontology import Ontology
    from .term import Term, TermList
    from .relationship import Relationship
except ImportError:
    pass






# coding: utf-8
"""
Test doctest contained tests in every file of the module.
"""

import os
import sys
import doctest
import shutil
import re
import warnings
import os.path as op


# Make sure we're using the local pronto library
sys.path.insert(0, op.dirname(op.dirname(op.abspath(__file__))))
import pronto

MODULE_TYPE = type(sys)
RESOURCES_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'resources')


class IgnoreUnicodeChecker(doctest.OutputChecker):
    def check_output(self, want, got, optionflags):
        if sys.version_info[0] > 2:
            want = re.sub("u'(.*?)'", "'\\1'", want)
            want = re.sub('u"(.*?)"', '"\\1"', want)
        return doctest.OutputChecker.check_output(self, want, got, optionflags)


def _load_tests_from_module(tests, module, globs, setUp, tearDown):
    """Load tests from module, iterating through submodules"""
    for attr in (getattr(module, x) for x in dir(module) if not x.startswith('_')):
        if isinstance(attr, MODULE_TYPE):
            tests.addTests(doctest.DocTestSuite(attr, globs=globs,
                setUp=setUp, tearDown=tearDown, checker=IgnoreUnicodeChecker()))
    return tests

def load_tests(loader, tests, ignore):
    """load_test function used by unittest to find the doctests"""

    def _setUp(self):
        """setUp method used by the DocTestSuite"""
        self._starting_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
        self._rundir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'run')
        os.mkdir(self._rundir)

    def _tearDown(self):
        """tearDown method used by the DocTestSuite"""
        os.chdir(self._starting_dir)
        shutil.rmtree(self._rundir)

    globs = {
        # ontologies
        'nmr': pronto.Ontology('http://nmrml.org/cv/v1.0.rc1/nmrCV.owl', False),
        'ms':  pronto.Ontology('https://raw.githubusercontent.com/HUPO-PSI/psi-ms-CV/master/psi-ms.obo'),
        'uo':  pronto.Ontology(os.path.join(RESOURCES_DIR, "uo.obo"), False),
        'cl':  pronto.Ontology(os.path.join(RESOURCES_DIR, "cl.ont"), False),

        # classes
        'Relationship': pronto.relationship.Relationship,
        'Ontology':     pronto.ontology.Ontology,
        'Term':         pronto.term.Term,
        'TermList':     pronto.term.TermList,
    }

    tests = _load_tests_from_module(tests, pronto, globs, _setUp, _tearDown)
    return tests



def setUpModule():
    warnings.simplefilter('ignore')

def tearDownModule():
    warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
"""
pronto.utils
============

This module contains some functions that are used in different parts
of the pronto library, as well as the definition of ProntoWarning.

Todo:
    * Maybe add a ProntoError class ?
"""

import six


class ProntoWarning(Warning):
    """A warning raised by pronto.

    Example:
        >>> from pronto import Ontology
        >>> import warnings
        >>> with warnings.catch_warnings(record=True) as w:
        ...    # the following ontology always has import issues (no URI in imports)
        ...    ims = Ontology('https://raw.githubusercontent.com/beny/imzml'
        ...                   '/master/data/imagingMS.obo')
        >>> print(w[-1].category)
        <class 'pronto.utils.ProntoWarning'>

    """
    pass

def unique_everseen(iterable):
    """List unique elements, preserving order. Remember all elements ever seen."""
    # unique_everseen('AAAABBBCCDAABBB')    --> A B C D
    seen = set()
    seen_add = seen.add

    for element in six.moves.filterfalse(seen.__contains__, iterable):
        seen_add(element)
        yield element

def output_str(f):
    if six.PY2:
        def new_f(*args, **kwargs):
            return f(*args, **kwargs).encode("utf-8")
    else:
        new_f = f
    return new_f



# coding: utf-8

import sys
import six
import os
import unittest
import warnings

from .utils import mock, py2skip, ciskip

# Make sure we're using the local pronto library
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import pronto


class TestProntoUtils(unittest.TestCase):

    def test_output_str(self):
        """Test function output_str provided by pronto.utils
        """
        @pronto.utils.output_str
        def unicode_returning_function():
            """This function returns str in Py3 and unicode in Py2"""
            return six.u("abc")

        self.assertIsInstance(unicode_returning_function(), str)

    def test_unique_everseen(self):
        """Test function unique_everseen provided by pronto.utils
        """
        self.assertEqual(['a', 'b', 'c', 'd'], list(pronto.utils.unique_everseen("aaaabbbcccaaadd")))

    def test_owlparser_get_id_from_url(self):
        """Test method _get_id_from_url of pronto.parser.owl.OwlXMLParser
        """
        test_values = {
            'http://purl.obolibrary.org/obo/IAO_0000115': 'IAO:0000115',
            'http://purl.obolibrary.org/obo/CL_0002420': 'CL:0002420',
            'http://purl.obolibrary.org/obo/UBERON_0001278': 'UBERON:0001278',
            'http://nmrML.org/nmrCV#NMR:1000001': 'NMR:1000001',
        }
        for k,v in six.iteritems(test_values):
            self.assertEqual(pronto.parser.owl.OwlXMLParser._get_id_from_url(k), v)


class TestTestUtils(unittest.TestCase):

    def setUp(self):
        warnings.simplefilter('ignore')

    def tearDown(self):
        warnings.simplefilter('default')

    def test_ciskip(self):
        """Test the ciskip function of the tests.utils module

        If there is an environment variable called CI with value true
        (as defined in a CI build environment), then the decorated function
        will not even be declared, and a mock function will be created and
        called instead.
        """
        for v,r in ('true', list('test')), ('false', list('estt')):
            with mock.patch.dict('os.environ', {'CI': v}):
                @ciskip
                def sort(l):
                    l.sort()
                l = list("test")
                sort(l)
                self.assertEqual(l, r)

    def test_py2skip(self):
        """Test the py2skip function of the tests.utils module

        If the python version is not 3, then the decorated function
        will not even be declared, and a mock function will be created and
        called instead.
        """
        for v,r in ((2,7,12), list('test')), ((3,5,2), list('estt')):
            with mock.patch('sys.version_info', v):
                @py2skip
                def sort(l):
                    l.sort()
                l = list("test")
                sort(l)
                self.assertEqual(l, r)


def setUpModule():
    warnings.simplefilter('ignore')

def tearDownModule():
    warnings.simplefilter(warnings.defaultaction)



# coding: utf-8
"""
pronto.parser.obo
=================

This module defines the Obo parsing method.
"""

import collections
import six

from .              import Parser
from .utils         import OboSection
from ..relationship import Relationship
from ..term         import Term


class OboParser(Parser):

    extensions = (".obo", ".obo.gz")

    def hook(self, **kwargs):
        """Returns True if this parser should be used.

        The current behaviour relies on filenames and file extension
        (.obo), but this is subject to change.
        """
        if 'force' in kwargs and kwargs['force']:
            return True
        if 'path' in kwargs:
            return kwargs['path'].endswith(self.extensions)

    @classmethod
    def parse(cls, stream):
        """Parse the stream.

        Parameters:
            stream (file handle): a binary stream of the file to parse

        Returns:
            dict: a dictionary containing the metadata headers
            dict: a dictionnary containing the terms
            set:  a set containing the imports

        """
        _section    = OboSection.meta
        meta        = collections.defaultdict(list)
        _rawterms   = []
        _rawtypedef = []

        for streamline in stream:

            # manage encoding && cleaning of line
            streamline = streamline.strip().decode('utf-8')
            if not streamline:
                continue

            _section = cls._check_section(streamline, _section)
            if _section is OboSection.meta:
                meta = cls._parse_metadata(streamline, meta)
            elif _section is OboSection.typedef:
                _rawtypedef = cls._parse_typedef(streamline, _rawtypedef)
            elif _section is OboSection.term:
                _rawterms = cls._parse_term(streamline, _rawterms)

        terms = cls._classify(_rawtypedef, _rawterms)
        imports = set(meta['import']) if 'import' in meta else set()

        return dict(meta), terms, imports

    @staticmethod
    def _check_section(line, section):
        """Updates the section currently parsed

        The parser starts in the OboSection.meta section but once
        it reaches the first [Typedef], it will enter the OboSection.typedef
        section, and/or when it reaches the first [Term], it will enter
        the OboSection.term section.
        """
        if line=="[Term]":
            section = OboSection.term
        if line=="[Typedef]":
            section = OboSection.typedef
        return section

    @classmethod
    def _parse_metadata(cls, line, meta, parse_remarks=True):
        """Parse a metadata line

        The metadata is organized as a "key: value" statement which
        is split into the proper key and the proper value.

        Parameters:
            line (str): the line containing the metadata
            parse_remarks(bool): if the remarks should be parsed as well
                (read the note) [default: True]

        Note:
            If the line follows the following schema: "remark: key: value",
            the function will attempt to extract the proper key/value
            instead of leaving everything inside the remark key.

            This may cause issues when the line is identified as such even
            though the remark is simply a sentence containing a colon, such as
            "remark: 090506 "Attribute" in Term deleted and new entries: Scan Type [...]"
            (found in imagingMS.obo). To prevent the splitting from happening,
            the text on the left of the colon must be less that *20 chars long*.
        """
        key, value = (x.strip() for x in line.split(':', 1))
        if parse_remarks and key=="remark":
            if 0<value.find(': ')<20:                                # Checking that the ':' is not
                try:                                                 # not too far avoid parsing a sentence
                    meta = cls._parse_metadata(value.strip(),        # containing a ':' as a key: value
                                               meta, parse_remarks)  # obo statement nested in a remark
                    return meta                                      # (20 is arbitrary, it may require
                except ValueError:                                   # tweaking)
                    pass
        meta[key].append(value)
        return meta

    @staticmethod
    def _parse_typedef(line, _rawtypedef):
        """Parse a typedef line

        The typedef is organized as a succesion of key:value pairs
        that are extracted into the same dictionnary until a new
        "[Typedef]" header is encountered

        Parameters:
            line (str): the line containing a typedef statement
        """
        if line.strip()=="[Typedef]":
            _rawtypedef.append(collections.defaultdict(list))
        else:
            key, value = (x.strip() for x in line.split(':', 1))
            _rawtypedef[-1][key].append(value)
        return _rawtypedef

    @staticmethod
    def _parse_term(line, _rawterms):
        """Parse a term line

        The term is organized as a succesion of key:value pairs
        that are extracted into the same dictionnary until a new
        "[Term]" header is encountered

        Parameters:
            line (str): the line containing a term statement
        """
        if line.strip()=="[Term]":
            _rawterms.append(collections.defaultdict(list))
        else:
            key, value = (x.strip() for x in line.split(':', 1))
            _rawterms[-1][key].append(value)
        return _rawterms

    @staticmethod
    def _classify(_rawtypedef, _rawterms):
        """Create proper objects out of the extracted dictionnaries

        New Relationship objects are instantiated with the help of
        the :obj:`Relationship._from_obo_dict` alternate constructor.

        New Term objects are instantiated by manually extracting id,
        name, desc and relationships out of the raw _term dictionnary,
        and then calling the default constructor.
        """
        terms = {}

        for _typedef in _rawtypedef:
            Relationship._from_obo_dict( # instantiate a new Relationship
                {k:v for k,lv in six.iteritems(_typedef) for v in lv}
            )

        for _term in _rawterms:
            _id   = _term['id'][0]
            try:
                _name = _term['name'][0]
            except IndexError:
                _name = ''
            finally:
                del _term['name']
            try:
                _desc = _term['def'][0]
            except IndexError:
                _desc = ''
            finally:
                del _term['def']
            _relations = collections.defaultdict(list)
            try:
                for other in _term['is_a']:
                    _relations[Relationship('is_a')].append(other.split('!')[0].strip())
            except IndexError:
                pass
            finally:
                del _term['is_a']
            try:
                for relname, other in ( x.split(' ', 1) for x in _term['relationship'] ):
                    _relations[Relationship(relname)].append(other.split('!')[0].strip())
            except IndexError:
                pass
            finally:
                del _term['relationship']
            terms[_id] = Term(_id, _name, _desc, dict(_relations), dict(_term))
        return terms


OboParser()



# coding: utf-8
"""
pronto.parser
=============

This module defines the Parser virtual class.
"""

__all__ = ["Parser", "OboParser", "OwlXMLParser"]


class Parser(object):
    """An abstract parser object.
    """

    _instances = {}

    def __init__(self):
        self._instances[type(self).__name__] = self

    #@pronto.utils.timeout(0)
    def parse(self, stream):
        """
        Parse the ontology file.

        Parameters
            stream (io.StringIO): A stream of the ontology file.

        Returns:
            dict: contains the metadata
            dict: contains the terms
            list: contains the imports
        """
        raise NotImplementedError


from .obo import OboParser
from .owl import OwlXMLParser, OwlXMLTreeParser, OwlXMLTargetParser



# coding: utf-8
from __future__ import absolute_import

### DEPS
import six
import unittest
import io
import sys
import contextlib
import os
import shutil
import gzip
import os.path as op
import warnings
import textwrap

from .utils import mock

if os.getcwd().endswith("pronto"):
    os.chdir("tests")

# Make sure we're using the local pronto library
sys.path.insert(0, op.dirname(op.dirname(op.abspath(__file__))))
import pronto


### TESTS
class TestProntoOntology(unittest.TestCase):

    def assert_loaded(self, ontology):

        # check if not empty
        self.assertNotEqual(len(ontology), 0)

        # check if succesfully crosslinked
        for term in ontology:
            for l in term.relations.values():
                for other in l:
                    self.assertIsInstance(other, pronto.Term)

    def assert_exportable(self, ontology):
        try:
            file = six.StringIO()
            file.write(ontology.obo)
        except BaseException as e:
            self.fail("export failed: {}".format(e))

    def assert_mergeable(self, ontology):
        try:
            other = pronto.Ontology()
            other.merge(ontology)
        except BaseException as e:
            self.fail("merge failed: {}".format(e))
        self.assertEqual(len(other), len(ontology))

    def check_ontology(self, ontology):
        self.assert_loaded(ontology)
        self.assert_exportable(ontology)
        self.assert_mergeable(ontology)





class TestProntoFeatures(TestProntoOntology):

    @classmethod
    def setUpClass(cls):
        cls.consistency_span = 10
        cls._rundir = os.path.join(os.getcwd(), 'run')
        os.mkdir(cls._rundir)

        ims_url = "https://github.com/beny/imzml/raw/master/data/imagingMS.obo"
        with open(os.path.join(cls._rundir, 'imagingMS.obo'), 'wb') as out_file:
            with contextlib.closing(six.moves.urllib.request.urlopen(ims_url)) as con:
                while True:
                    chunk = con.read(1024)
                    if not chunk:
                        break
                    out_file.write(chunk)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls._rundir)

    def test_ims_consistency(self):
        """Assert several runs on the same file give the same output.

        See [https://github.com/althonos/pronto/issues/4]
        Thanks to @winni-genp for issue reporting.
        """
        obo = pronto.Ontology(os.path.join(self._rundir, "imagingMS.obo"), False)
        expected_keys = set(obo.terms.keys())
        for _ in six.moves.range(self.consistency_span):
            tmp_obo = pronto.Ontology(os.path.join(self._rundir, "imagingMS.obo"), False)
            self.assertEqual(set(tmp_obo.terms.keys()), expected_keys)
        self.check_ontology(obo)

    def test_remote_obogz_parsing(self):
        """Test the behaviour of pronto when parsing remote gzipped files.

        On Python2, this should raise a NotImplementedError (as
        urllib2.urlopen does not implement :obj:`io.StringIO` methods,
        gzip.GzipFile cannot handle on-the-fly decoding of data).

        On Python3, this should actually produce a proper ontology.
        """
        hpo_url = "https://github.com/Bioconductor-mirror/gwascat/"\
                      "raw/master/inst/obo/hpo.obo.gz"
        if six.PY3:
            gzipped = pronto.Ontology(hpo_url)
            expected_keys = {
                x.decode('utf-8').strip().split(': ')[-1]
                    for x in gzip.GzipFile(
                        fileobj=six.moves.urllib.request.urlopen(hpo_url)
                    )
                        if x.startswith(b'id: ')
            }
            self.assertEqual(expected_keys, set(gzipped.terms.keys()))
            self.check_ontology(gzipped)
        else:
            with self.assertRaises(NotImplementedError):
                gzipped = pronto.Ontology(hpo_url, parser='OwlXMLTargetParser')

    def test_unicode_in_term_names(self):
        """Test if unicode characters in term names work.

        See [https://github.com/althonos/pronto/issues/6]
        Thanks to @owen-jones-gen for issue reporting.
        """
        ontology = pronto.Ontology("resources/owen-jones-gen.obo")
        with mock.patch('sys.stdout', new=six.moves.StringIO()) as output:
            for term in ontology:
                print(term)
                print(term.obo)
            print(ontology.obo)

            self.assertEqual(output.getvalue().strip(),
                             textwrap.dedent("""
                                             <ONT0:ROOT: °>
                                             [Term]
                                             id: ONT0:ROOT
                                             name: °
                                             [Term]
                                             id: ONT0:ROOT
                                             name: °
                                             """).strip())

        self.check_ontology(ontology)

    def test_exportable_headerless_ontology(self):
        """Test if a headerless ontology can be exported

        See [https://github.com/althonos/pronto/issues/5]
        Thanks to @winni-genp for issue reporting.
        """
        obo = pronto.Ontology("resources/winni-genp.obo")
        self.check_ontology(obo)

    def test_with_custom_typedef(self):
        """Try to import an obo ontology containing custom typedefs
        """
        obo = pronto.Ontology("resources/elo.obo")
        self.check_ontology(obo)

        self.assertIn("has_written", pronto.Relationship._instances)
        self.assertIn(pronto.Relationship("written_by"), pronto.Relationship._instances.values())
        self.assertIn(pronto.Relationship('written_by'), obo['ELO:0130001'].relations)
        self.assertIn(pronto.Relationship('has_written'), obo['ELO:0330001'].relations)


class TestProntoLocalOntology(TestProntoOntology):

    def test_local_owl_noimports(self):
        """Try to import a local owl ontology without its imports
        """
        owl = pronto.Ontology("resources/cl.ont", False)
        self.check_ontology(owl)

    def test_local_obo_noimports(self):
        """Try to import a local obo ontology without its imports
        """
        obo = pronto.Ontology("resources/cmo.obo", False)
        self.check_ontology(obo)

    def test_local_owl_imports(self):
        """Try to import a local owl ontology with its imports
        """
        owl = pronto.Ontology("resources/cl.ont")
        self.check_ontology(owl)

    def test_local_obo_imports(self):
        """Try to import a local owl ontology with its imports
        """
        obo = pronto.Ontology("resources/cmo.obo")
        self.check_ontology(obo)


class TestProntoRemoteOntology(TestProntoOntology):

    def test_remote_obo_noimports(self):
        """Try to import a remote obo ontology without its imports
        """
        obo = pronto.Ontology("http://purl.obolibrary.org/obo/pdumdv.obo", False)
        self.check_ontology(obo)

    def test_remote_owl_noimports(self):
        """Try to import a remote owl ontology without its imports
        """
        owl = pronto.Ontology("http://aber-owl.net/onts/FLU_63.ont", False)
        self.check_ontology(owl)

    def test_remote_obo_imports(self):
        """Try to import a remote obo ontology with its imports
        """
        obo = pronto.Ontology("http://purl.obolibrary.org/obo/doid.obo")
        self.check_ontology(obo)

    def test_remote_owl_imports(self):
        """Try to import a remote owl ontology without its imports
        """
        owl = pronto.Ontology("http://purl.obolibrary.org/obo/xao.owl")
        self.check_ontology(owl)


def setUpModule():
    warnings.simplefilter('ignore')

def tearDownModule():
    warnings.simplefilter(warnings.defaultaction)
